module bestsed

! Author: Stefan Noll (2009) 

! Writes the best-fitting models of objects in 'outspec/*.spec' files.
! Needs the corresponding '*.X2b' and 'outspec/e*.spec' model files.

contains 

subroutine bests



IMPLICIT NONE

CHARACTER(LEN=*), PARAMETER :: filenames = 'cigale.par' ! INPUT!
CHARACTER(LEN=*), PARAMETER :: cosmofile = 'cosmology.dat' ! INPUT!
INTEGER, PARAMETER :: ncol = 6 ! OUTPUT!
REAL(KIND=8), PARAMETER :: minflux = 1. ! in W/A
REAL(KIND=8), PARAMETER :: c_dist = 1d-5 ! Mpc -> 10 pc
REAL(KIND=8), PARAMETER :: c_area = 1.1967d36 ! surface at 10 pc in m^2
REAL(KIND=8), PARAMETER :: c_flamfnu = 1d19 ! W/(m^2*A) -> mJy
REAL(KIND=8), PARAMETER :: c = 2.99792458d8 ! light speed in m/s

CHARACTER(LEN=80) :: outdir, infilelist, photdat, photdatfnu, suffix 
CHARACTER(LEN=80) :: photdatX2b, specfile, specout, sedout
CHARACTER(LEN=160) :: header
CHARACTER(LEN=32) :: ID
CHARACTER(LEN=1) :: sharp, cdummy
CHARACTER(LEN=800) :: catparlab, bmparlab, bmparval, catparval
INTEGER :: h, nfiles, idummy, ndustmod, nlam, ndustmod0, nlam0, i, nrow, nfilt 
INTEGER :: npar, nobj, iobj, nbest, j, file0, dustmod, row0, ifiles, idustmod
INTEGER :: status_open, status_read 
REAL(KIND=8) :: rdummy, mz, mgal, H0, Omega_M, Omega_L, d_L

REAL(KIND=8), DIMENSION(:), ALLOCATABLE :: lam, lam0, nflux, flux, zlam
REAL(KIND=8), DIMENSION(:), ALLOCATABLE :: zflux, fluxcorr, czflux, flam, fnu


! *** Read file names ***

OPEN (10, STATUS='OLD', FILE=filenames, IOSTAT=status_open)
IF (status_open /= 0) THEN
   PRINT*, 'Parameter file ', filenames, ' not present!'
   STOP
END IF

DO h = 1, 24
   SELECT CASE (h)
   CASE (2) 
      READ (10, '(A)') outdir
      IF (INDEX(outdir, '/', .TRUE.) /= LEN(TRIM(outdir))) &
           outdir = TRIM(outdir) // '/' 
   CASE (6) 
      READ (10, '(A)') infilelist
      infilelist = TRIM(infilelist) // '.list'
   CASE (22)
      READ (10, '(A)') photdat
      photdatfnu = TRIM(photdat) // '.fnu'
   CASE (24)
      READ (10, '(A)') suffix
      IF (suffix(1:2) == '_ ') suffix = ' '
      photdatX2b = TRIM(photdat) // TRIM(suffix) // '.X2b'
   CASE (4,8,10,12,14,16,18,20)
      READ (10, '(A)') header
   CASE DEFAULT
      READ (10, '(A)') header
      IF (header(1:1) /= '#') PRINT*, 'Wrong header!'
   END SELECT
END DO

CLOSE (10)


! *** Read cosmological parameters ***

OPEN (20, STATUS='OLD', FILE=cosmofile, IOSTAT=status_open)
IF (status_open /= 0) THEN
   PRINT*, TRIM(cosmofile), ' is missing!'
   STOP
END IF

READ (20, '(A)') header
READ (20, *) H0, Omega_M, Omega_L

CLOSE (20)


! *** Prepare reading of spectra ***

OPEN (30, STATUS='OLD', FILE=infilelist, IOSTAT=status_open)
IF (status_open /= 0) THEN
   PRINT*, TRIM(infilelist), ' is missing!'
   STOP
END IF

nfiles = 0 
DO 

   READ (30, '(A)', IOSTAT=status_read) specfile
   IF (status_read /= 0) EXIT
   nfiles = nfiles + 1

   OPEN (40, STATUS='OLD', FILE=specfile, IOSTAT=status_open)
   IF (status_open /= 0) THEN
      PRINT*, 'File ', TRIM(specfile), ' not present!'
      STOP
   END IF

   READ (40, '(A)') header  
   READ (40, *) sharp, idummy, idummy, idummy, ndustmod, nlam

   IF (nfiles > 1) THEN
      IF (ndustmod /= ndustmod0) THEN
         PRINT*, 'Number of dust models do not match!'
         STOP
      END IF
      IF (nlam /= nlam0) THEN
         PRINT*, 'Number of wavelengths do not match!'
         STOP
      END IF
   END IF
   ndustmod0 = ndustmod
   nlam0 = nlam

   READ (40, '(A)') header
   READ (40, '(A)') header

   ! Read wavelengths

   IF (nfiles == 1) ALLOCATE (lam(nlam), lam0(nlam))

   READ (40, '(A)') header
   READ (40, *) (lam(i), i = 1, nlam)

   IF (nfiles > 1 .AND. .NOT.ALL(lam == lam0)) THEN
      PRINT*, 'Wavelengths of models do not match!'
      STOP
   END IF
   lam0 = lam

   CLOSE (40)
 
END DO

CLOSE (30)

nrow = INT(nlam / ncol)
IF (nrow * ncol < nlam) nrow = nrow + 1

ALLOCATE (nflux(nlam), flux(nlam))
ALLOCATE (zlam(nlam), zflux(nlam))
ALLOCATE (fluxcorr(nlam), czflux(nlam))
ALLOCATE (flam(nlam), fnu(nlam))


! *** Open catalogue file with filter fluxes ***

OPEN (50, STATUS='OLD', FILE=photdatfnu, IOSTAT=status_open)
IF (status_open /= 0) THEN
   PRINT*, 'File ', TRIM(photdatfnu), ' not present!'
   STOP
END IF
READ (50, '(A)') header
READ (50, '(A)') header
READ (50, '(A)') catparlab


! *** Open file with properties of the best-fitting models ***

OPEN (60, STATUS='OLD', FILE=photdatX2b, IOSTAT=status_open)
IF (status_open /= 0) THEN
   PRINT*, 'File ', TRIM(photdatX2b), ' not present!'
   STOP
END IF
READ (60, '(A)') header
READ (60, *) sharp, idummy, nfilt, npar, nobj
READ (60, '(A)') bmparlab


! *** Identify best-fitting models and write spectra ***

DO iobj = 1, nobj

   ! Read object data
   READ (50, '(A)') catparval

   ! Read object IDs and best-model numbers
   READ (60, *) ID, nbest, mz, (rdummy, j = 1, nfilt+1), mgal 
   BACKSPACE (60)
   READ (60, '(A)') bmparval

   ! Determine file and position of best model
   file0 = INT((nbest - 1) / ndustmod)
   dustmod = MODULO((nbest - 1), ndustmod) + 1
   row0 = 5 + (3 + nrow) * dustmod

   ! Read fluxes

   OPEN (30, STATUS='OLD', FILE=infilelist)
   DO ifiles = 1, file0
      READ (30, '(A1)') cdummy 
   END DO
   READ (30, '(A)') specfile 
   CLOSE (30)

   OPEN (40, STATUS='OLD', FILE=specfile)
   DO idustmod = 1, row0
      READ (40, '(A)') cdummy 
   END DO
   READ (40, *) (nflux(i), i = 1, nlam)
   CLOSE (40)

   ! The galaxy mass scales the models: 
   flux = nflux * 10 ** mgal
   WHERE (flux < minflux) flux = 0.

   ! Effect of redshift
   zlam = lam * (1 + mz)
   zflux = flux / (1 + mz) 
 
   ! Consider absorption by the IGM
   CALL CorrIGM (zlam, mz, fluxcorr)
   czflux = zflux * fluxcorr

   ! f_nu at luminosity distance
   d_L = LumDist(mz, H0, Omega_M, Omega_L)
   IF (d_L < c_dist) d_L = c_dist
   flam = (czflux / c_area) * (c_dist / d_L) ** 2

   ! f_lam in f_nu
   fnu = (c_flamfnu / c) * (zlam ** 2) * flam

   ! Write spectra

   specout = TRIM(outdir) // TRIM(photdat) // TRIM(suffix) // '_' // &
        TRIM(ID) // '.spec'
   sedout = TRIM(outdir) // TRIM(photdat) // TRIM(suffix) // '_' // &
        TRIM(ID) // '.sed'

   PRINT*, "Model: ", TRIM(specout), "& SED: ", TRIM(sedout)

   OPEN (70, STATUS='REPLACE', FILE=specout)

   WRITE (70, '(A)') '# Nfilt Npar Nlam'
   WRITE (70, '(A1,1X,I2,1X,I2,1X,I4)') '#', nfilt, npar, nlam
   WRITE (70, '(A)') TRIM(catparlab)
   WRITE (70, '(A1,1X,A)') '#', TRIM(catparval)
   WRITE (70, '(A)') TRIM(bmparlab)
   WRITE (70, '(A1,1X,A)') '#', TRIM(bmparval)
   WRITE (70, '(A)') '# lam[A] Flam[W/A] zlam[A] zflam[W/m2A] zfnu[mJy]'
   WRITE (70, '(E11.5,1X,E10.4,1X,E11.5,1X,E10.4,1X,E10.4)') & 
        (lam(i), flux(i), zlam(i), flam(i), fnu(i), i = 1, nlam) 

   CLOSE (70)

END DO
 
CLOSE (50)
CLOSE (60)


CONTAINS

  SUBROUTINE CorrIGM (lam, z, fluxcorr)

    ! IGM absorption according to Meiksin (2006)

    REAL(KIND=8), DIMENSION(:), INTENT(IN) :: lam ! observed wavelength
    REAL(KIND=8), INTENT(IN) :: z ! redshift
    REAL(KIND=8), DIMENSION(:), INTENT(OUT) :: fluxcorr ! corr(lambda)

    INTEGER, PARAMETER :: nmax = 31, mmax = 10
    REAL(KIND=8), PARAMETER :: lamL = 912 ! Lyman limit in A
    REAL(KIND=8), DIMENSION(2:9), PARAMETER :: fac = &
         (/1., 0.348, 0.179, 0.109, 0.0722, 0.0508, 0.0373, 0.0283/)
    REAL(KIND=8), PARAMETER :: gamma = 0.2788 ! Gamma(0.5,1)

    INTEGER :: n, m
    REAL(KIND=8) :: taua, term1, term2
    REAL(KIND=8), DIMENSION(2:nmax) :: lamn
    REAL(KIND=8), DIMENSION(2:nmax,SIZE(lam)) :: zn, taun
    REAL(KIND=8), DIMENSION(SIZE(lam)) :: zL, tauL_IGM, term3, term4
    REAL(KIND=8), DIMENSION(SIZE(lam)) :: tauL_LLS, tau 

    DO n = 2, nmax
       lamn(n) = lamL / (1. - 1. / n ** 2)
       zn(n,:) = lam / lamn(n) - 1.
    END DO

    IF (z >= 4) THEN
       taua = 0.00058 * (1. + z) ** 4.5
    ELSE
       taua = 0.00211 * (1. + z) ** 3.7
    END IF

    IF (z >= 4) THEN
       taun(2,:) = 0.00058 * (1. + zn(2,:)) ** 4.5
    ELSE
       taun(2,:) = 0.00211 * (1. + zn(2,:)) ** 3.7
    END IF 
    DO n = 3, 5
       WHERE (zn(n,:) < 3)
          taun(n,:) = taua * fac(n) * (0.25 * (1. + zn(n,:))) ** (1. / 3.)
       ELSEWHERE
          taun(n,:) = taua * fac(n) * (0.25 * (1. + zn(n,:))) ** (1. / 6.)
       END WHERE
    END DO
    DO n = 6, 9 
       taun(n,:) = taua * fac(n) * (0.25 * (1. + zn(n,:))) ** (1. / 3.)
    END DO
    DO n = 10, nmax
       taun(n,:) = taun(9,:) * 720. / (n * (n ** 2 - 1.)) 
    END DO
    WHERE (zn >= z) taun = 0.

    zL = lam / lamL - 1.
    WHERE (zL >= z)
       tauL_IGM = 0.
    ELSEWHERE
       tauL_IGM = 0.805 * (1. + zL) ** 3 * (1. / (1. + zL) - 1. / (1. + z))
    END WHERE

    term1 = gamma - EXP(-1.)
    term2 = 0
    DO m = 0, mmax-1
       term2 = term2 + (-1.) ** m / (rfacul(m) * REAL((2 * m - 1), 8))
    END DO
    term3 = (1. + z) * (lam / lamL) ** 1.5 - (lam / lamL) ** 2.5
    term4 = 0.
    DO m = 1, mmax
       term4 = term4 + (2. * (-1.) ** m / &
            (rfacul(m) * REAL((2 * m - 1) * (6 * m - 5), 8))) * &
            ((1. + z) ** (2.5 - REAL(3 * m, 8)) * (lam / lamL) ** (3 * m) - &
            (lam / lamL) ** 2.5)
    END DO
    WHERE (zL >= z)
       tauL_LLS = 0.      
    ELSEWHERE
       tauL_LLS = 0.25 * ((term1 - term2) * term3 - term4) 
    END WHERE

    tau = SUM(taun, DIM=1) + tauL_IGM + tauL_LLS 
    fluxcorr = EXP(- tau)

  END SUBROUTINE CorrIGM


  REAL(KIND=8) FUNCTION rfacul(n)

    ! Computes real faculty from integer value

    INTEGER, INTENT(IN) :: n
    INTEGER :: imax, i

    IF (n == 0) THEN
       imax = 1
    ELSE
       imax = n 
    END IF

    rfacul = 1.
    DO i = 1, imax
      rfacul = rfacul * i 
    END DO

  END FUNCTION rfacul


  REAL(KIND=8) FUNCTION LumDist(z, H0, Omega_M, Omega_L)

    ! Computes luminosity distance at redshift z in Mpc for given Lambda 
    ! cosmology (H_0 in (km/s)/Mpc, Omega_M, and Omega_Lambda) 
    ! Ref.: Correll et al. (1992) / Riess et al. (1998)

    IMPLICIT NONE
    REAL(KIND=8) :: z
    REAL(KIND=8), OPTIONAL :: H0, Omega_M, Omega_L
    REAL(KIND=8), PARAMETER :: H0_sc = 70, Omega_M_sc = .3, Omega_L_sc = .7
    REAL(KIND=8), PARAMETER :: tol = 1d-12
!    REAL(KIND=8), PARAMETER :: dz = 1d-3 ! accuracy of integration
    INTEGER, PARAMETER :: nstep = 1001 ! define the accuracy instead of dz
    REAL(KIND=8), PARAMETER :: c = 2.99792458d5 ! light speed in km/s
    INTEGER :: imax, i
    REAL(KIND=8) :: dl, dzi, zi, Omega_k

    IF (.NOT. PRESENT(H0)) H0 = H0_sc
    IF (.NOT. PRESENT(Omega_M)) Omega_M = Omega_M_sc
    IF (.NOT. PRESENT(Omega_L)) Omega_L = Omega_L_sc

    dl = 0
    IF (z > tol) THEN
!       imax = NINT(z / dz)
       imax = nstep
       dzi = z / imax
       DO i = 1, imax 
!          zi = (i - 0.5) * dz
          zi = (i - 0.5) * dzi
          dl = dl + dzi * ((1 + zi) ** 2 * (1 + Omega_M * zi) - &
               zi * (2 + zi) * Omega_L) ** (-0.5)
       END DO
    ELSE IF (z < -tol) THEN
      PRINT*, 'LumDist: z < 0 -> Assume z = 0!'
      z = 0 
    END IF

    Omega_k = 1 - Omega_M - Omega_L
    IF (ABS(Omega_k) > tol) THEN
       dl = dl * (ABS(Omega_k) ** 0.5)
       IF (Omega_k > 0) THEN
          dl = SINH(dl)
       ELSE IF (Omega_k < 0) THEN
          dl = SIN(dl)
       END IF
       dl = dl * (ABS(Omega_k) ** (-0.5))
    END IF

    LumDist = c * (H0 ** (-1)) * (1 + z) * dl

  END FUNCTION LumDist
end subroutine bests
END module bestsed
