module prepagnspec

! Author: Stefan Noll (2009) 

! Preparation of Siebenmorgen et al. AGN models for the use in CIGALE

! Command line parameters:
! (1) file name of AGN model ('agnx_*')
! (2) file name for wavelength grid (optional, default: 'inspec/irdh01.spec')



contains 

  subroutine agn


IMPLICIT NONE

CHARACTER(LEN=80), PARAMETER :: reffile0 = 'inspec/irdh01.spec'
INTEGER, PARAMETER :: nin = 236
REAL(KIND=8), PARAMETER :: llim = 912. ! in A
REAL(KIND=8), PARAMETER :: ulim = 1.6e7 ! in A

CHARACTER(LEN=80) :: file, infile, outfile, reffile
INTEGER :: narg, ndot, status_open, i, nout
REAL(KIND=8) :: minflux, rdummy, lum
REAL(KIND=8), DIMENSION(nin) :: inlam, influx, inlamc, influxc
REAL(KIND=8), DIMENSION(nin) :: lginlam, lginflux

REAL(KIND=8), DIMENSION(:), ALLOCATABLE :: outlam, lgoutlam, lgoutflux
REAL(KIND=8), DIMENSION(:), ALLOCATABLE :: outfluxc, outflux


! Get command line parameters

narg = iargc()

IF (narg == 0) THEN

   PRINT*, 'No AGN model given! -> Stop!'
   STOP  

ELSE

   CALL GETARG(1, file)
   ndot = INDEX(file, '.plot')
   IF (ndot /= 0) file = TRIM(file(1:ndot-1))
   infile = TRIM(file) // '.plot'
   outfile = TRIM(file) // '.spec'

   IF (narg >= 2) THEN
      CALL GETARG(2, reffile)
      ndot = INDEX(reffile, '.spec')
      IF (ndot == 0) reffile = TRIM(reffile) // '.spec'
   ELSE
      reffile = reffile0
      PRINT*, 'Take ', TRIM(reffile), ' for wavelength grid.'  
   END IF

END IF


! Read AGN model

OPEN (10, STATUS='OLD', FILE=infile, IOSTAT=status_open)
IF (status_open /= 0) THEN
   PRINT*, TRIM(infile), ' is missing!'
   STOP
END IF

READ (10, *) (inlam(i), influx(i), i = nin, 1, -1)  

CLOSE (10)


! Change units of AGN model

! mum -> A
inlamc = inlam * 1e4 
! Jy -> W/(m^2A)
influxc = 300. * influx / inlamc ** 2

! avoid 0 values
minflux = MINVAL(influxc, influxc > 0)
WHERE (influxc == 0.) influxc = minflux

! in log
lginlam = LOG10(inlamc)
lginflux = LOG10(influxc) 


! Read reference wavelengths 

OPEN (20, STATUS='OLD', FILE=reffile, IOSTAT=status_open)
IF (status_open /= 0) THEN
   PRINT*, TRIM(reffile), ' is missing!'
   STOP
END IF

READ (20, *) nout
ALLOCATE (outlam(nout))
READ (20, *) (outlam(i), rdummy, i = 1, nout)

CLOSE (20)

! in log
ALLOCATE (lgoutlam(nout))
lgoutlam = LOG10(outlam)


! Adapt wavelength grid of AGN model

ALLOCATE (lgoutflux(nout))
CALL InExpolCurve (lginlam, lginflux, lgoutlam, lgoutflux)
WHERE (lgoutflux == 0.) lgoutflux = -100  


! Flux in output units

ALLOCATE (outfluxc(nout))
outfluxc = 10 ** lgoutflux
WHERE (outfluxc < 1d-99) outfluxc = 0.


! Normalise AGN model

ALLOCATE (outflux(nout))
lum = IntegrateRange(outlam, outfluxc, llim, ulim)
outflux = outfluxc / lum


! Write adapted AGN model into file 

OPEN (30, STATUS='REPLACE', FILE=outfile)

WRITE (30, '(I4)') nout
DO i = 1, nout
 WRITE (30, '(E11.5,1X,E10.4)') outlam(i), outflux(i)
END DO

CLOSE (30)


CONTAINS

  SUBROUTINE InExpolCurve (inx, infx, refx, outfx)

    ! Derive function values for a different set of data points as given for
    ! the function

    IMPLICIT NONE    
    REAL(KIND=8), DIMENSION(:), INTENT(IN) :: inx, infx, refx
    REAL(KIND=8), DIMENSION(:), INTENT(OUT) :: outfx
    INTEGER :: ninx, nrefx, j, i
    INTEGER, DIMENSION(1) :: linx
    REAL(KIND=8) :: minx, maxx

    ninx = SIZE(inx)
    minx = inx(1)
    maxx = inx(ninx)

    nrefx = SIZE(refx)
    IF (refx(1) > minx) THEN
       linx = MINLOC(inx, inx > refx(1))
       j = linx(1)
    ELSE
       j = 2
    END IF

    DO i = 1, nrefx 

       IF (refx(i) < minx) THEN
          outfx(i) = &
               interplinlin(inx(1), inx(2), infx(1), infx(2), refx(i))
       ELSE IF (refx(i) > maxx) THEN
          outfx(i) = interplinlin(inx(ninx-1), inx(ninx), &
               infx(ninx-1), infx(ninx), refx(i))
       ELSE IF (refx(i) == maxx) THEN
          outfx(i) = infx(ninx)
       ELSE
          DO WHILE (refx(i) < inx(j-1) .OR. refx(i) >= inx(j))
             j = j + 1
          END DO
          outfx(i) = &
               interplinlin(inx(j-1), inx(j), infx(j-1), infx(j), refx(i))
       END IF

    END DO

  END SUBROUTINE InExpolCurve


  REAL(KIND=8) FUNCTION interplinlin(x1, x2, y1, y2, t)
    
    IMPLICIT NONE
    REAL(KIND=8) :: x1, x2, y1, y2, t
    
    interplinlin = y1 + (y2 - y1) * (t - x1) / (x2 - x1)
    
  END FUNCTION interplinlin


  REAL(KIND=8) FUNCTION IntegrateRange(lambda, flux, llim, ulim)

    ! Integrates curve in a given range

    IMPLICIT NONE
    REAL(KIND=8), DIMENSION(:), INTENT(IN) :: lambda, flux
    REAL(KIND=8), INTENT(IN) :: llim, ulim
    INTEGER :: imax, i
    REAL(KIND=8) :: llam, ulam, dlambda

    imax = MIN(SIZE(lambda), SIZE(flux))
    IF (imax < SIZE(lambda)) PRINT*, &
         'IntegrateRange: More wavelengths than fluxes!'
    IF (imax > SIZE(lambda)) PRINT*, & 
         'IntegrateRange: Less wavelengths than fluxes!' 

    IntegrateRange = 0.

    IF (ulim <= llim) THEN

       PRINT*, 'IntegrateRange: Upper limit <= lower limit!'
     
    ELSE

       DO i = 1, imax

          IF (i == 1) THEN
             llam = lambda(i) - (lambda(i+1) - lambda(i)) / 2
          ELSE
             llam = (lambda(i) + lambda(i-1)) / 2
          END IF

          IF (i == imax) THEN
             ulam = lambda(i) + (lambda(i) - lambda(i-1)) / 2
          ELSE
             ulam = (lambda(i+1) + lambda(i)) / 2
          END IF
             
          IF (llim <= ulam .AND. ulim >= llam) THEN

             IF (llim > llam .OR. ulim < ulam) THEN
                IF (llim > llam) dlambda = MAX(ulam - llim, 0.)
                IF (ulim < ulam) dlambda = MAX(ulim - llam, 0.)
             ELSE                
                dlambda = ulam - llam
             END IF

             IntegrateRange = IntegrateRange + dlambda * flux(i)

          END IF

       END DO

    END IF

  END FUNCTION IntegrateRange

  end subroutine agn

END module prepagnspec
