program prepphotdat

! Author: Stefan Noll (2009) 

! Prepares photometric data for the chi^2 fitting by calculating f_nu 
! from magitudes (and lg(L_IR/L_sol) if desired)


IMPLICIT NONE

CHARACTER(LEN=*), PARAMETER :: filenames = 'cigale.par' ! INPUT!
CHARACTER(LEN=*), PARAMETER :: cosmofile = 'cosmology.dat' ! INPUT! 
INTEGER, PARAMETER :: nmaxlamfilt = 1500 ! Check 'filterfile'!
REAL(KIND=8), PARAMETER :: tol = 1d-10
REAL(KIND=8), PARAMETER :: c_dist = 1d-5 ! Mpc -> 10 pc
REAL(KIND=8), PARAMETER :: Lsol = 3.827d26 ! in W
REAL(KIND=8), PARAMETER :: meanlam_IR = 1d6 ! in A (arbitrary)
REAL(KIND=8), PARAMETER :: c_area = 1.1967d36 ! surface at 10 pc in m^2
REAL(KIND=8), PARAMETER :: c_flamfnu = 1d19 ! W/(m^2*A) -> mJy
REAL(KIND=8), PARAMETER :: c = 2.99792458d8 ! light speed in m/s

CHARACTER(LEN=80) :: filterfile, calibfile, filtoutpar
CHARACTER(LEN=80) :: photdat, photdatmag, photdatfnu
CHARACTER(LEN=800) :: header, fnupar, allpar, fmt 
CHARACTER(LEN=32) :: ID
CHARACTER(LEN=8) :: cdummy
CHARACTER(LEN=1) :: sharp, sign
INTEGER :: h, nfilters, i, j, idummy, m, m_IR, mmax
INTEGER :: hmax, nval, n, nmax, isel
INTEGER :: status_open, status_read,zacc
REAL(KIND=8) :: H0, Omega_M, Omega_L, z, d_L, mag0, flam0, fnu0
REAL(KIND=8) :: relerr,systerr

CHARACTER(LEN=8), DIMENSION(:), ALLOCATABLE :: filtlab
CHARACTER(LEN=16), DIMENSION(:), ALLOCATABLE :: valarr
INTEGER, DIMENSION(:), ALLOCATABLE :: nlamfilt, typetrans
INTEGER, DIMENSION(:), ALLOCATABLE :: ifilt, typecalib
REAL(KIND=8), DIMENSION(:,:), ALLOCATABLE :: lamfilt, trans
REAL(KIND=8), DIMENSION(:), ALLOCATABLE :: fluxVega, area, meanlam, ABVega
REAL(KIND=8), DIMENSION(:), ALLOCATABLE :: calib, mag, dmag, fnu, dfnu


! *** Read file names ***

OPEN (10, STATUS='OLD', FILE=filenames, IOSTAT=status_open)
IF (status_open /= 0) THEN
   PRINT*, 'Parameter file ', filenames, ' not present!'
   STOP
END IF

DO h = 1, 16
   SELECT CASE (h)
   CASE (2) 
      READ (10, '(A)') filterfile
      filterfile = TRIM(filterfile) // '.dat'
   CASE (4) 
      READ (10, '(A)') calibfile
      calibfile = TRIM(calibfile) // '.dat'
   CASE (6) 
      READ (10, '(A)') filtoutpar
      filtoutpar = TRIM(filtoutpar) // '.par'
   CASE (10)
      READ (10, '(A)') photdat
      photdatmag = TRIM(photdat) // '.mag'
      photdatfnu = TRIM(photdat) // '.fnu'
CASE (12)
   READ (10, *) zacc
   CASE(14)
   READ (10, *) relerr
   CASE(16)
      READ (10, *) systerr 

      CASE (8)
      READ (10, '(A)') header   
   CASE DEFAULT
      READ (10, '(A)') header
      IF (header(1:1) /= '#') PRINT*, 'Wrong header!'
   END SELECT
END DO

CLOSE (10)


! *** Read the filter transmissions ***

OPEN (20, STATUS='OLD', FILE=filterfile, IOSTAT=status_open)
IF (status_open /= 0) THEN
   PRINT*, TRIM(filterfile), ' is missing!'
   STOP
END IF

READ (20, '(A)') header
READ (20, *) sharp, nfilters
READ (20, '(A)') header
READ (20, '(A)') header

ALLOCATE (nlamfilt(nfilters))
ALLOCATE (typetrans(nfilters))
ALLOCATE (lamfilt(nfilters,nmaxlamfilt))
ALLOCATE (trans(nfilters,nmaxlamfilt))

DO i = 1, nfilters
   READ (20, *) nlamfilt(i), typetrans(i)
   READ (20, *) (lamfilt(i,j), trans(i,j), j = 1, nlamfilt(i))
   IF (nlamfilt(i) > nmaxlamfilt) &
        PRINT*, 'Warning: Filter ', i, 'not completely read!'
END DO

CLOSE (20)


! *** Read the filter calibrations ***

ALLOCATE (fluxVega(nfilters))
ALLOCATE (area(nfilters))
ALLOCATE (meanlam(nfilters))
ALLOCATE (ABVega(nfilters))

OPEN (30, STATUS='OLD', FILE=calibfile, IOSTAT=status_open)
IF (status_open /= 0) THEN
   PRINT*, TRIM(calibfile), ' is missing!'
   STOP
END IF

READ (30, '(A)') header
DO i = 1, nfilters
   READ (30, *) idummy, cdummy, fluxVega(i), area(i), meanlam(i), ABVega(i)
END DO

CLOSE(30)


! *** Read output information for filter fluxes ***

ALLOCATE (ifilt(nfilters))
ALLOCATE (filtlab(nfilters))
ALLOCATE (typecalib(nfilters))
filtlab = ' '

OPEN (40, STATUS='OLD', FILE=filtoutpar, IOSTAT=status_open)
IF (status_open /= 0) THEN
   PRINT*, TRIM(filtoutpar), ' is missing!'
   STOP
END IF

DO
   READ (40, '(A)', IOSTAT=status_read) header
   IF (header(1:1) /= '#') EXIT
END DO
BACKSPACE (40)

m = 0
m_IR = 0
DO 
   m = m + 1
   READ (40, *, IOSTAT=status_read) &
        ifilt(m), filtlab(m), typecalib(m), cdummy
   IF (status_read /= 0 .OR. filtlab(m) == ' ') EXIT
   IF (ifilt(m) == 0) m_IR = m
END DO
mmax = m - 1

CLOSE (40)


! *** Consider selected magnitude system ***

ALLOCATE (calib(mmax))

DO m = 1, mmax
   SELECT CASE (typecalib(m))
   CASE (0) ! flux
      calib(m) = 0.
   CASE (1) ! ref = Vega
      calib(m) = 2.5 * LOG10(fluxVega(ifilt(m))) + 0.03
   CASE (2) ! ref = AB
      calib(m) = 2.5 * LOG10(fluxVega(ifilt(m))) + ABVega(ifilt(m))
   CASE DEFAULT
      PRINT*, 'No valid calibration system selected! -> Assume Vega!'
      calib(m) = 2.5 * LOG10(fluxVega(ifilt(m))) + 0.03
   END SELECT
END DO


! *** Read cosmological parameters ***

OPEN (50, STATUS='OLD', FILE=cosmofile, IOSTAT=status_open)
IF (status_open /= 0) THEN
   PRINT*, TRIM(cosmofile), ' is missing!'
   STOP
END IF

READ (50, '(A)') header
READ (50, *) H0, Omega_M, Omega_L

CLOSE (50)


! *** Read magnitudes and write filter fluxes ***

! Required structure of data file: 
! Optional line of column labels starting with '#'
! ID[string] z[0.-10.] (mmax X mag[0.-100.] [, dmag])
! mmax from 'filtoutpar'

ALLOCATE (mag(mmax), dmag(mmax))
ALLOCATE (fnu(mmax), dfnu(mmax))

OPEN (60, STATUS='OLD', FILE=photdatmag, IOSTAT=status_open)
IF (status_open /= 0) THEN
   PRINT*, TRIM(photdatmag), ' is missing!'
   STOP
END IF

h = 0
DO
   READ (60, '(A)', IOSTAT=status_read) header
   IF (status_read /= 0) THEN
      PRINT*, TRIM(photdatmag), ' does not contain data! -> Stop!'
      STOP
   END IF
   IF (header(1:1) /= '#') EXIT
   h = h + 1
END DO
hmax = h
BACKSPACE (60)

! Magnitudes with or without errors?

ALLOCATE (valarr(2+mmax*2))
CALL SplitString (TRIM(header), valarr)
IF (LEN(TRIM(valarr(3+mmax))) == 0) THEN 
   nval = 1
ELSE
   nval = 2
END IF

! Derive number of objects

n = 0
DO
   ID = ' '
   READ (60, '(A)', IOSTAT=status_read) ID
   IF (status_read /= 0 .OR. ID == ' ') EXIT
   n = n + 1 
END DO
nmax = n
REWIND (60)

! Write output file header

OPEN (70, STATUS='REPLACE', FILE=photdatfnu)

WRITE (70, '(A)') '# Nval Nfilt Nobj'
WRITE (70, '(A1,1X,I1,1X,I2,1X,I5)') '#', nval, mmax, nmax

fnupar = ' '
DO m = 1, mmax
   fnupar = TRIM(fnupar) // ' f' // TRIM(filtlab(m)) // '[mJy] df' // &
        TRIM(filtlab(m)) // '[mJy]'
END DO
allpar = '# ID z' // TRIM(fnupar)
WRITE (70, '(A)') TRIM(allpar)

! Read magnitudes and write fluxes

fmt = '(A,1X,F8.6' // REPEAT(',1X,E9.4', 2*mmax) // ')'

DO h = 1, hmax
   READ (60, '(A)') header
END DO

DO n = 1, nmax
  
   ID = ' '
   IF (nval == 1) THEN
      READ (60, *) ID, z, (mag(m), m = 1, mmax)
      dmag = 0.
   ELSE
      READ (60, *) ID, z, (mag(m), dmag(m), m = 1, mmax)
   END IF
   PRINT*, ID 

   IF (z < 0. - tol .OR. z > 9. - tol) THEN
      PRINT*, 'Invalid redshift -> Assume z = 9.9'
      z = 9.9
   END IF

   IF (m_IR > 0) THEN
      ! Luminosity distance in Mpc for z:
      d_L = LumDist(z, H0, Omega_M, Omega_L)
      IF (d_L < c_dist) d_L = c_dist
   END IF

   DO m = 1, mmax

      IF (typecalib(m) == 0 .AND. m /= m_IR) THEN

         IF (mag(m) < 0. + tol) THEN
            PRINT*, 'Invalid f', TRIM(filtlab(m)), ' -> Assume f', &
                 TRIM(filtlab(m)), ' = 0'
            mag(m) = 0.
         END IF
         IF (nval == 2 .AND. dmag(m) < 0. + tol) THEN
            PRINT*, 'Invalid df', TRIM(filtlab(m)), ' -> Assume df', &
                  TRIM(filtlab(m)), ' = 0'
            dmag(m) = 0.
         END IF

      ELSE IF (m == m_IR) THEN

         IF (mag(m) < 0. + tol .OR. mag(m) > 99. - tol) THEN
            PRINT*, 'Invalid L_IR -> Assume L_IR = 0'
            mag(m) = 0.
         END IF
         IF (nval == 2 .AND. & 
              (dmag(m) < 0. + tol .OR. dmag(m) > 99. - tol)) THEN
            PRINT*, 'Invalid dL_IR -> Assume dL_IR = 0'
            dmag(m) = 0.
         END IF

      ELSE

         IF (mag(m) < 0. + tol .OR. mag(m) > 99. - tol) THEN
            PRINT*, 'Invalid magnitude ', TRIM(filtlab(m)), &
                 ' -> Assume mag = 99.99'
            mag(m) = 99.99
         END IF
         IF (nval == 2 .AND. &
              (dmag(m) < 0. + tol .OR. dmag(m) > 99. - tol)) THEN
            PRINT*, 'Invalid magnitude error d', TRIM(filtlab(m)), &
                 ' -> Assume mag = 0.'
            dmag(m) = 0.
         END IF

      END IF

   END DO
        
   DO m = 1, mmax

      IF (typecalib(m) == 0 .AND. m /= m_IR) THEN

         fnu(m) = mag(m)
         dfnu(m) = dmag(m)

      ELSE IF (m == m_IR) THEN

         ! Error column contains optional upper limit!
         IF (mag(m) < tol .AND. dmag(m) > tol) THEN
            mag0 = dmag(m) 
         ELSE
            mag0 = mag(m)
         END IF

         ! Calculate f_nu from lg(L_IR/L_sol) assuming arbitrary meanlam_IR
         flam0 = (10 ** mag0 * Lsol / meanlam_IR) / (c_area * (1 + z)) 
         fnu0 = (c_flamfnu / c) * (meanlam_IR ** 2) * flam0
         fnu0 = fnu0 / (d_L / c_dist) ** 2

         ! Error column contains optional upper limit!
         IF (mag(m) < tol .AND. dmag(m) > tol) THEN
            dfnu(m) = fnu0
            fnu(m) = 0.
         ELSE
            fnu(m) = fnu0
            ! Approximate mag +- dmag -> fnu +- dfnu
            dfnu(m) = SINH(LOG(10.) * dmag(m)) * fnu(m) 
         END IF

      ELSE

         ! Error column contains optional upper limit!
         IF (mag(m) > 99. - tol .AND. dmag(m) > tol) THEN
            mag0 = dmag(m) 
         ELSE
            mag0 = mag(m)
         END IF

         fnu0 = (c_flamfnu / c) * (meanlam(ifilt(m)) ** 2) * &
              10 ** (-0.4 * (mag0 - calib(m)))    

         ! Error column contains optional upper limit!
         IF (mag(m) > 99. - tol .AND. dmag(m) > tol) THEN
            dfnu(m) = fnu0
            fnu(m) = 0.
         ELSE
            fnu(m) = fnu0
            ! Approximate mag +- dmag -> fnu +- dfnu
            dfnu(m) = SINH(0.4 * LOG(10.) * dmag(m)) * fnu(m) 
         END IF

      END IF

      IF (fnu(m) < tol) fnu(m) = 0.
      IF (dfnu(m) < tol) dfnu(m) = 0.

   END DO

   WRITE (70, fmt) TRIM(ID), z, (fnu(m), dfnu(m), m = 1, mmax)

END DO

CLOSE (60)
CLOSE (70)


CONTAINS

  SUBROUTINE SplitString (instr, outstrarr)

    ! Splits a string at the positions of inner spaces.

    IMPLICIT NONE
    CHARACTER(LEN=*), INTENT(IN) :: instr
    CHARACTER(LEN=*), DIMENSION(:), INTENT(OUT) :: outstrarr   
    CHARACTER(LEN=LEN(instr)) :: tmpstr
    INTEGER :: i

    outstrarr = ' '
    tmpstr = TRIM(ADJUSTL(instr)) 
    IF (LEN(tmpstr) /= 0) THEN
       DO i = 1, SIZE(outstrarr)
          outstrarr(i) = TRIM(tmpstr(1:INDEX(tmpstr,' ')-1)) 
          tmpstr = TRIM(ADJUSTL(tmpstr(INDEX(tmpstr,' '):))) 
          IF (LEN(tmpstr) == 0) EXIT 
       END DO
    END IF

  END SUBROUTINE SplitString


  REAL(KIND=8) FUNCTION LumDist(z, H0, Omega_M, Omega_L)

    ! Computes luminosity distance at redshift z in Mpc for given Lambda 
    ! cosmology (H_0 in (km/s)/Mpc, Omega_M, and Omega_Lambda) 
    ! Ref.: Correll et al. (1992) / Riess et al. (1998)

    IMPLICIT NONE
    REAL(KIND=8) :: z
    REAL(KIND=8), OPTIONAL :: H0, Omega_M, Omega_L
    REAL(KIND=8), PARAMETER :: H0_sc = 70, Omega_M_sc = .3, Omega_L_sc = .7
    REAL(KIND=8), PARAMETER :: tol = 1d-12
    REAL(KIND=8), PARAMETER :: dz = 1d-3 ! accuracy of integration
    REAL(KIND=8), PARAMETER :: c = 2.99792458d5 ! light speed in km/s
    INTEGER :: imax, i
    REAL(KIND=8) :: dl, dzi, zi, Omega_k

    IF (.NOT. PRESENT(H0)) H0 = H0_sc
    IF (.NOT. PRESENT(Omega_M)) Omega_M = Omega_M_sc
    IF (.NOT. PRESENT(Omega_L)) Omega_L = Omega_L_sc

    dl = 0
    IF (z > tol) THEN
       imax = NINT(z / dz)
       dzi = z / imax
       DO i = 1, imax 
          zi = (i - 0.5) * dz
          dl = dl + dzi * ((1 + zi) ** 2 * (1 + Omega_M * zi) - &
               zi * (2 + zi) * Omega_L) ** (-0.5)
       END DO
    ELSE IF (z < -tol) THEN
      PRINT*, 'LumDist: z < 0 -> Assume z = 0!'
      z = 0 
    END IF

    Omega_k = 1 - Omega_M - Omega_L
    IF (ABS(Omega_k) > tol) THEN
       dl = dl * (ABS(Omega_k) ** 0.5)
       IF (Omega_k > 0) THEN
          dl = SINH(dl)
       ELSE IF (Omega_k < 0) THEN
          dl = SIN(dl)
       END IF
       dl = dl * (ABS(Omega_k) ** (-0.5))
    END IF

    LumDist = c * (H0 ** (-1)) * (1 + z) * dl

  END FUNCTION LumDist

END program prepphotdat
